@@ -1,12 +1,12 @@
 package Test::BDD::Cucumber::Executor;
-$Test::BDD::Cucumber::Executor::VERSION = '0.59';
+$Test::BDD::Cucumber::Executor::VERSION = '0.61';
 =head1 NAME
 
 Test::BDD::Cucumber::Executor - Run through Feature and Harness objects
 
 =head1 VERSION
 
-version 0.59
+version 0.61
 
 =head1 DESCRIPTION
 
@@ -17,51 +17,24 @@ Definitions, and reporting on progress through the passed-in harness.
 
 use Moo;
 use MooX::HandlesVia;
-use Types::Standard qw( Bool ArrayRef HashRef );
+use Types::Standard qw( Bool Str ArrayRef HashRef );
 use Clone qw(clone);
-use List::Util qw/first/;
+use List::Util qw/first any/;
 use List::MoreUtils qw/pairwise/;
 use Module::Runtime qw/use_module/;
 use Number::Range;
 use Carp qw/croak/;
 our @CARP_NOT;
 
-use Test::Builder;
+use Test2::API qw/intercept context/;
 
-# Setup wrapping for Test::Builder
-use Test::BDD::Cucumber::TestBuilderDelegator;
-use Devel::Refcount qw/refcount/;
-if ( ( !$ENV{'TEST_BDD_CUCUMBER_NO_TB_WRAP_TEST'} )
-    && refcount($Test::Builder::Test) > 1 )
-{
+# Use-ing the formatter results in a
+# 'loaded too late to be used globally' warning
+# But we only need it locally anyway.
+require Test2::Formatter::TAP;
 
-    my $ref_trace = "[Install Devel::FindRef to see these diagnostics]";
-    if ( eval { use_module "Devel::FindRef" } ) {
-        $ref_trace = Devel::FindRef::track($Test::Builder::Test);
-    }
-
-    my $message = sprintf( <<'END', $ref_trace );
-!!! HEY YOU !!!
-Test::BDD::Cucumber needs to be able to wrap $Test::Builder::Test in order to
-properly capture testing output. However, something else has already taken a
-reference to that module. You need to `use` Test::BDD::Cucumber::Executor
-before the other testing modules are used. Modules that appear to already have a
-reference are:
------
-%s
------
-You can safetly ignore the global $Test::Builder::Test. In almost all cases, the
-simple fix is the move the line that says `use Test::BDD::Cucumber::Executor`
-above all other `use Test::*` lines. You can also suppress this check by setting:
-
-  TEST_BDD_CUCUMBER_NO_TB_WRAP_TEST=1
-END
-
-    croak $message;
-}
-
-$Test::Builder::Test =
-  Test::BDD::Cucumber::TestBuilderDelegator->new( Test::Builder->new() );
+# Needed for subtest() -- we don't want to import all its functions though
+require Test::More;
 
 use Test::BDD::Cucumber::StepContext;
 use Test::BDD::Cucumber::Util;
@@ -70,6 +43,22 @@ use Test::BDD::Cucumber::Errors qw/parse_error_from_line/;
 
 has '_bail_out' => ( is => 'rw', isa => Bool, default => 0 );
 
+=head1 ATTRIBUTES
+
+=head2 matching
+
+The value of this attribute should be one of C<first> (default), C<relaxed> and C<strict>.
+
+By default (C<first>), the first matching step is executed immediately,
+terminating the search for (further) matching steps. When C<matching> is set
+to anything other than C<first>, all steps are checked for matches. When set
+to C<relaxed>, a warning will be generated on multiple matches. When set to
+C<strict>, an exception will be thrown.
+
+=cut
+
+has matching => ( is => 'rw', isa => Str, default => 'first');
+
 =head1 METHODS
 
 =head2 extensions
@@ -111,7 +100,13 @@ C<add_steps()> takes step definitions of the item list form:
   [ Given => qr//, sub {} ],
  ),
 
-and populates C<steps> with them.
+Or, when metadata is specified with the step, of the form:
+
+ (
+  [ Given => qr//, { meta => $data }, sub {} ]
+ ),
+
+(where the hashref stores step metadata) and populates C<steps> with them.
 
 =cut
 
@@ -122,27 +117,30 @@ sub add_steps {
 
     # Map the steps to be lower case...
     for (@steps) {
-        my ( $verb, $match, $code ) = @$_;
+        my ( $verb, $match, $meta, $code );
+
+        if (@$_ == 3) {
+            ( $verb, $match, $code ) = @$_;
+            $meta = {};
+        }
+        else {
+            ( $verb, $match, $meta, $code ) = @$_;
+        }
         $verb = lc $verb;
 
-        if ( $verb =~ /^(before|after)$/ ) {
-            $code  = $match;
-            $match = qr//;
-        } else {
-            unless ( ref($match) ) {
-                $match =~ s/:\s*$//;
-                $match = quotemeta($match);
-                $match = qr/^$match:?/i;
-            }
+        unless ( ref($match) ) {
+            $match =~ s/:\s*$//;
+            $match = quotemeta($match);
+            $match = qr/^$match:?/i;
         }
 
         if ( $verb eq 'transform' or $verb eq 'after' ) {
 
             # Most recently defined Transform takes precedence
             # and After blocks need to be run in reverse order
-            unshift( @{ $self->{'steps'}->{$verb} }, [ $match, $code ] );
+            unshift( @{ $self->{'steps'}->{$verb} }, [ $match, $meta, $code ] );
         } else {
-            push( @{ $self->{'steps'}->{$verb} }, [ $match, $code ] );
+            push( @{ $self->{'steps'}->{$verb} }, [ $match, $meta, $code ] );
         }
 
     }
@@ -470,7 +468,7 @@ sub add_table_placeholders {
 
 Accepts a L<Test::BDD::Cucumber::StepContext> object, and searches through
 the steps that have been added to the executor object, executing against the
-first matching one.
+first matching one (unless C<$self->matching> indicates otherwise).
 
 You can also pass in a boolean 'short-circuit' flag if the Scenario's remaining
 steps should be skipped, and a boolean flag to denote if it's a redispatched
@@ -487,9 +485,39 @@ sub find_and_dispatch {
       if $short_circuit;
 
     # Try and find a matching step
-    my $step = first { $context->text =~ $_->[0] }
-    @{ $self->{'steps'}->{ $context->verb } || [] },
-      @{ $self->{'steps'}->{'step'} || [] };
+    my $step;
+    my $text = $context->text;
+    if ($self->matching eq 'first') {
+        $step = first { $text =~ $_->[0] }
+        @{ $self->{'steps'}->{ $context->verb } || [] },
+            @{ $self->{'steps'}->{'step'} || [] };
+    }
+    else {
+        my @steps = grep { $text =~ $_->[0] }
+        @{ $self->{'steps'}->{ $context->verb } || [] },
+            @{ $self->{'steps'}->{'step'} || [] };
+
+        if (@steps > 1) {
+            my $filename = $context->step->line->document->filename;
+            my $line = $context->step->line->number;
+            my $msg =
+                join("\n   ",
+                     qq(Step "$text" ($filename:$line) matches multiple step functions:),
+                     map {
+                       qq{matcher $_->[0] defined at } .
+                           (($_->[1]->{source} && $_->[1]->{line})
+                            ? "$_->[1]->{source}:$_->[1]->{line}"
+                            : '<unknown>') } @steps);
+
+            if ($self->matching eq 'relaxed') {
+                warn $msg;
+            }
+            else {
+                die $msg;
+            }
+        }
+        $step = shift @steps;
+    }
 
     # Deal with the simple case of no-match first of all
     unless ($step) {
@@ -527,28 +555,7 @@ sub dispatch {
       if $short_circuit;
 
     # Execute the step definition
-    my ( $regular_expression, $coderef ) = @$step;
-
-    # Setup what we'll pass to step_done, with out localized Test::Builder
-    # stuff
-    my $output    = '';
-    my $tb_return = {
-        output  => \$output,
-        builder => Test::Builder->create()
-    };
-
-    # Set its outputs to be self-referential
-    $tb_return->{'builder'}->output( \$output );
-    $tb_return->{'builder'}->failure_output( \$output );
-    $tb_return->{'builder'}->todo_output( \$output );
-
-    binmode($tb_return->{'builder'}->output(), ':utf8');
-    binmode($tb_return->{'builder'}->failure_output(), ':utf8');
-    binmode($tb_return->{'builder'}->todo_output(), ':utf8');
-
-    # Make a minimum pass
-    $tb_return->{'builder'}
-      ->ok( 1, "Starting to execute step: " . $context->text );
+    my ( $regular_expression, $meta, $coderef ) = @$step;
 
     my $step_name = $redispatch ? 'sub_step' : 'step';
     my $step_done_name = $step_name . '_done';
@@ -559,64 +566,69 @@ sub dispatch {
     # Store the string position of matches for highlighting
     my @match_locations;
 
-    # New scope for the localization
-    my $result;
     my $stash_keys = join ';', sort keys %{$context->stash};
-    {
-        # Localize test builder
-        local $Test::Builder::Test->{'_wraps'} = $tb_return->{'builder'};
+    # Using `intercept()`, run the step function in an isolated
+    # environment -- this should not affect the enclosing scope
+    # which might be a TAP::Harness scope.
+    #
+    # Instead, we want the tests inside this scope to map to
+    # status values
+    my $events = intercept {
+        # This is a hack to make Test::More's $TODO variable work
+        # inside the intercepted scope.
 
-        no warnings 'redefine';
-        local *Test::Builder::BAIL_OUT = sub {
-            my ( $tb, $message ) = @_;
-            $self->_bail_out(1);
-            local @CARP_NOT = qw(Test::More Test::BDD::Cucumber::Executor);
-            croak("BAIL_OUT() called: $message");
-        };
+        ###TODO: Both intercept() and Test::More::subtest() should
+        # be replaced by a specific Hub implementation for T::B::C
+        Test::More::subtest( 'execute step', sub {
+            my $ctx = context();
+            $ctx->pass( "Starting to execute step: " . $context->text );
 
-        # Execute!
+            # Execute!
+            no warnings 'redefine';
 
-        # Set S and C to be step-specific values before executing the step
-        local *Test::BDD::Cucumber::StepFile::S = sub {
-            return $context->stash->{'scenario'};
-        };
-        local *Test::BDD::Cucumber::StepFile::C = sub {
-            return $context;
-        };
+            local *Test::BDD::Cucumber::StepFile::S = sub {
+                return $context->stash->{'scenario'};
+            };
+            local *Test::BDD::Cucumber::StepFile::C = sub {
+                return $context;
+            };
 
-        # Take a copy of this. Turns out actually matching against it
-        # directly causes all sorts of weird-ass heisenbugs which mst has
-        # promised to investigate.
-        my $text = $context->text;
+            # Take a copy of this. Turns out actually matching against it
+            # directly causes all sorts of weird-ass heisenbugs which mst has
+            # promised to investigate.
+            my $text = $context->text;
 
-        # Save the matches
-        $context->matches( [ $text =~ $regular_expression ] );
+            # Save the matches
+            $context->matches( [ $text =~ $regular_expression ] );
 
-        # Save the location of matched subgroups for highlighting hijinks
-        my @starts = @-;
-        my @ends   = @+;
-        @match_locations = pairwise { [ $a, $b ] } @starts, @ends;
+            # Save the location of matched subgroups for highlighting hijinks
+            my @starts = @-;
+            my @ends   = @+;
+            @match_locations = pairwise { [ $a, $b ] } @starts, @ends;
 
-        # OK, actually execute
-        eval { $coderef->($context) };
-        if ($@) {
-            $Test::Builder::Test->ok( 0, "Test compiled" );
-            $Test::Builder::Test->diag($@);
-        }
-
-        # Close up the Test::Builder object
-        $tb_return->{'builder'}->done_testing();
-
-        my $status = $self->_test_status( $tb_return->{builder} );
-
-        # Create the result object
-        $result = Test::BDD::Cucumber::Model::Result->new(
-            {
-                result => $status,
-                output => $output
+            # OK, actually execute
+            eval { $coderef->($context) };
+            if ($@) {
+                $ctx->fail("Step ran successfully", $@);
             }
-        );
-    }
+
+            $ctx->done_testing;
+            $ctx->release;
+                             });
+    };
+
+    my $status = $self->_test_status( $events );
+
+    my $result = Test::BDD::Cucumber::Model::Result->new(
+        {
+            result => $status,
+            # due to the hack above with the subtest inside the
+            # interception scope, we need to grovel the subtest
+            # from out of the other results first.
+            output => $self->_test_output(
+                (first { $_->isa('Test2::Event::Subtest') }
+                 @$events)->{subevents})
+        });
     warn qq|Unsupported: Step modified C->stash instead of C->stash->{scenario} or C->stash->{feature}|
         if $stash_keys ne (join ';', sort keys %{$context->stash});
 
@@ -673,54 +685,64 @@ sub _extract_match_strings {
     return @parts;
 }
 
+sub _test_output {
+    my ($self, $events) = @_;
+    my $fmt = Test2::Formatter::TAP->new();
+    open my $stdout, '>', \my $out_text;
+    my $idx = 0;
+
+    $fmt->set_handles([ $stdout, $stdout ]);
+    $self->_test_output_from_subevents($events, $fmt, \$idx);
+    close $stdout;
+
+    return $out_text;
+}
+
+sub _test_output_from_subevents {
+    my ($self, $events, $fmt, $idx) = @_;
+
+    for my $event (@$events) {
+        if ($event->{subevents}) {
+            $self->_test_output_from_subevents(
+                $event->{subevents}, $fmt, $idx);
+        }
+        else {
+            $fmt->write($event, $$idx++);
+        }
+    }
+}
+
 sub _test_status {
     my $self    = shift;
-    my $builder = shift;
+    my $events  = shift;
 
-    my $results =
-        $builder->can("history")
-      ? $self->_test_status_from_history($builder)
-      : $self->_test_status_from_details($builder);
-
-    # Turn that in to a Result status
-    return
-        $results->{'fail'} ? 'failing'
-      : $results->{'todo'} ? 'pending'
-      :                      'passing';
+    if (any { defined $_->{effective_pass}
+              and ! $_->{effective_pass} } @$events) {
+        return 'failing';
+    }
+    else {
+        return $self->_test_status_from_subevents($events) ? 'pending' : 'passing';
+    }
 }
 
-sub _test_status_from_details {
+sub _test_status_from_subevents {
     my $self    = shift;
-    my $builder = shift;
+    my $events  = shift;
 
-    # Make a note of test status
-    my %results = map {
-        if ( $_->{'ok'} ) {
-            if ( $_->{'type'} eq 'todo' || $_->{'type'} eq 'todo_skip' ) {
-                ( todo => 1 );
-            } else {
-                ( pass => 1 );
-            }
-        } else {
-            ( fail => 1 );
+    for my $e (@$events) {
+        if (exists $e->{subevents}) {
+            $self->_test_status_from_subevents($e->{subevents})
+                and return 1;
         }
-    } $builder->details;
+        elsif (defined $e->{amnesty}
+               and $e->{effective_pass}
+               and (not $e->{pass})
+               and any { $_->{tag} eq 'TODO' } @{$e->{amnesty}}) {
+            return 1;
+        }
+    }
 
-    return \%results;
-}
-
-sub _test_status_from_history {
-    my $self    = shift;
-    my $builder = shift;
-
-    my $history = $builder->history;
-
-    my %results;
-    $results{todo} = $history->todo_count ? 1 : 0;
-    $results{fail} = !$history->test_was_successful;
-    $results{pass} = $history->pass_count ? 1 : 0;
-
-    return \%results;
+    return 0;
 }
 
 =head2 skip_step
